﻿/*
 * AttachedCurve.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.Runtime.Serialization;

using Boare.Lib.Vsq;

namespace Boare.Cadencii {

    /// <summary>
    /// VsqTrackの各カーブに付属させるBezierChainのリスト
    /// </summary>
    [Serializable]
    public class AttachedCurve : ICloneable {
        /// <summary>
        /// カーブ本体
        /// </summary>
        private Dictionary<int, BezierChain>[] m_curves;

        public object Clone() {
            AttachedCurve ret = new AttachedCurve();
            for ( int i = 0; i < m_curves.Length; i++ ) {
                ret.m_curves[i] = new Dictionary<int, BezierChain>();
                foreach ( int key in m_curves[i].Keys ) {
                    ret.m_curves[i].Add( key, (BezierChain)m_curves[i][key].Clone() );
                }
            }
            return ret;
        }

        public AttachedCurve() {
            m_curves = new Dictionary<int, BezierChain>[9];
            for ( int i = 0; i < 9; i++ ) {
                m_curves[i] = new Dictionary<int, BezierChain>();
            }
        }

        public AttachedCurve( BezierCurves item ) {
            m_curves = new Dictionary<int, BezierChain>[9];
            for ( int i = 0; i < 9; i++ ) {
                m_curves[i] = new Dictionary<int, BezierChain>();
            }
            int track = -1;
            
            track++;
            for ( int i = 0; i < item.Dynamics.Length; i++ ) {
                BezierChain chain = (BezierChain)item.Dynamics[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.Brethiness.Length; i++ ) {
                BezierChain chain = (BezierChain)item.Brethiness[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.Brightness.Length; i++ ) {
                BezierChain chain = (BezierChain)item.Brightness[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.Clearness.Length; i++ ) {
                BezierChain chain = (BezierChain)item.Clearness[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.Opening.Length; i++ ) {
                BezierChain chain = (BezierChain)item.Opening[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.GenderFactor.Length; i++ ) {
                BezierChain chain = (BezierChain)item.GenderFactor[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.PortamentoTiming.Length; i++ ) {
                BezierChain chain = (BezierChain)item.PortamentoTiming[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.PitchBend.Length; i++ ) {
                BezierChain chain = (BezierChain)item.PitchBend[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }

            track++;
            for ( int i = 0; i < item.PitchBendSensitivity.Length; i++ ) {
                BezierChain chain = (BezierChain)item.PitchBendSensitivity[i].Clone();
                for ( int j = 0; j < chain.points.Count; j++ ) {
                    chain.points[j].ID = j;
                }
                m_curves[track].Add( i, chain );
            }
        }

        public int GetNextId( VsqCurveType curve_type ) {
            int index = VsqCurveTypeUtil.IndexFromCurveType( curve_type );
            int ret = m_curves[index].Count;
            while ( m_curves[index].ContainsKey( ret ) ) {
                ret++;
            }
            return ret;
        }

        /// <summary>
        /// 指定したコントロールカーブにベジエ曲線を追加します。BezierChainのIDは自動で割り振られ、戻り値として返されます
        /// </summary>
        /// <param name="curve_type"></param>
        /// <param name="chain"></param>
        /// <returns></returns>
        public int AddBezierChain( VsqCurveType curve_type, BezierChain chain ) {
            int index = VsqCurveTypeUtil.IndexFromCurveType( curve_type );
            int ret = GetNextId( curve_type );
            m_curves[index].Add( ret, (BezierChain)chain.Clone() );
            return ret;
        }

        public void AddBezierChain( VsqCurveType curve_type, BezierChain chain, int chain_id ) {
            int index = VsqCurveTypeUtil.IndexFromCurveType( curve_type );
            m_curves[index].Add( chain_id, (BezierChain)chain.Clone() );
        }

        public Dictionary<int, BezierChain> this[ VsqCurveType vct ] {
            get {
                int index = VsqCurveTypeUtil.IndexFromCurveType( vct );
                if ( 0 <= index && index < m_curves.Length ) {
                    return m_curves[index];
                } else {
                    return null;
                }
            }
        }
    }

}
