﻿/*
 * BezierChain.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.Serialization;

namespace Boare.Cadencii {

    [Serializable]
    public class BezierChain : IDisposable, ICloneable {
        public List<BezierPoint> points;
        public int Default = 0;
        public int ID;
        private Color m_color;

        /// <summary>
        /// 与えられたBezierChainがx軸について陰かどうかを判定する
        /// </summary>
        /// <param name="chain"></param>
        /// <returns></returns>
        public static bool IsBezierImplicit( BezierChain chain ) {
            for ( int i = 0; i < chain.points.Count - 1; i++ ) {
                float pt1 = chain.points[i].Base.X;
                float pt2 = (chain.points[i].ControlRightType == BezierControlType.None) ? pt1 : chain.points[i].ControlRight.X;
                float pt4 = chain.points[i + 1].Base.X;
                float pt3 = (chain.points[i + 1].ControlLeftType == BezierControlType.None) ? pt4 : chain.points[i + 1].ControlLeft.X;
                if ( !IsUnitBezierImplicit( pt1, pt2, pt3, pt4 ) ) {
                    return false;
                }
            }
            return true;
        }

        /// <summary>
        /// 4つの制御点からなるベジエ曲線が、x軸について陰かどうかを判定する
        /// </summary>
        /// <param name="pt1">始点</param>
        /// <param name="pt2">制御点1</param>
        /// <param name="pt3">制御点2</param>
        /// <param name="pt4">終点</param>
        /// <returns></returns>
        private static bool IsUnitBezierImplicit( float pt1, float pt2, float pt3, float pt4 ) {
            double a = pt4 - 3.0 * pt3 + 3.0 * pt2 - pt1;
            double b = 2.0 * pt3 - 4.0 * pt2 + 2.0 * pt1;
            double c = pt2 - pt1;
            if ( a == 0.0f ) {
                if ( c >= 0.0f && b + c >= 0.0f ) {
                    return true;
                } else {
                    return false;
                }
            } else if ( a > 0.0f ) {
                if ( -b / (2.0f * a) <= 0.0f ) {
                    if ( c >= 0.0f ) {
                        return true;
                    } else {
                        return false;
                    }
                } else if ( 1.0f <= -b / (2.0f * a) ) {
                    if ( a + b + c >= 0.0f ) {
                        return true;
                    } else {
                        return false;
                    }
                } else {
                    if ( c - b * b / (4.0f * a) >= 0.0f ) {
                        return true;
                    } else {
                        return false;
                    }
                }
            } else {
                if ( -b / (2.0f * a) <= 0.5f ) {
                    if ( a + b + c >= 0.0f ) {
                        return true;
                    } else {
                        return false;
                    }
                } else {
                    if ( c >= 0.0f ) {
                        return true;
                    } else {
                        return false;
                    }
                }
            }
        }

        [OnDeserialized]
        private void onDeserialized( StreamingContext sc ) {
            for ( int i = 0; i < points.Count; i++ ) {
                points[i].ID = i;
            }
        }

        public void Dispose() {
            if ( points != null ) {
                points.Clear();
            }
        }

        public int GetNextId() {
            if ( points.Count > 0 ) {
                int max = points[0].ID;
                for ( int i = 1; i < points.Count; i++ ) {
                    max = Math.Max( max, points[i].ID );
                }
                return max + 1;
            } else {
                return 0;
            }
        }
        
        public void GetValueMinMax( out int min, out int max ) {
            //todo: ベジエが有効なときに、曲線の描く最大値、最小値も考慮
            min = Default;
            max = Default;
            foreach ( BezierPoint bp in points ) {
                min = Math.Min( min, bp.Base.Y );
                max = Math.Max( max, bp.Base.Y );
            }
        }

        public void GetKeyMinMax( out int min, out int max ) {
            min = Default;
            max = Default;
            foreach ( BezierPoint bp in points ) {
                min = Math.Min( min, bp.Base.X );
                max = Math.Max( max, bp.Base.X );
            }
        }
        
        public object Clone() {
            BezierChain result = new BezierChain( this.m_color );
            foreach ( BezierPoint bp in points ) {
                result.points.Add( (BezierPoint)bp.Clone() );
            }
            result.Default = this.Default;
            result.ID = ID;
            return result;
        }

        public BezierChain( Color curve ) {
            points = new List<BezierPoint>();
            m_color = curve;
        }

        public BezierChain() {
            points = new List<BezierPoint>();
            m_color = Color.Black;
        }

        public BezierPoint this[int index] {
            get {
                return points[index];
            }
            set {
                points[index] = value;
            }
        }
        
        public Color Color {
            get {
                return m_color;
            }
            set {
                m_color = value;
            }
        }

        public void Add( BezierPoint bp ) {
            if ( points == null ) {
                points = new List<BezierPoint>();
                m_color = Color.Black;
            }
            points.Add( bp );
            points.Sort();
        }

        public void RemoveAt( int id ) {
            for ( int i = 0; i < points.Count; i++ ) {
                if ( points[i].ID == id ) {
                    points.RemoveAt( i );
                    break;
                }
            }
        }

        public int Count {
            get {
                if ( points == null ) {
                    return 0;
                }
                return points.Count;
            }
        }

        public float GetValue( float x ) {
            for ( int i = 0; i < points.Count - 1; i++ ) {
                if ( points[i].Base.X <= x && x <= points[i + 1].Base.X ) {
                    float x1 = points[i].Base.X;
                    float x4 = points[i + 1].Base.X;
                    if ( x1 == x ) {
                        return points[i].Base.Y;
                    } else if ( x4 == x ) {
                        return points[i + 1].Base.Y;
                    } else {
                        float x2 = points[i].ControlRight.X;
                        float x3 = points[i + 1].ControlLeft.X;
                        float a3 = x4 - 3 * x3 + 3 * x2 - x1;
                        float a2 = 3 * x3 - 6 * x2 + 3 * x1;
                        float a1 = 3 * (x2 - x1);
                        float a0 = x1;
                        float t = SolveCubicEquation( a3, a2, a1, a0, x );
                        x1 = points[i].Base.Y;
                        x2 = points[i].ControlRight.Y;
                        x3 = points[i + 1].ControlLeft.Y;
                        x4 = points[i + 1].Base.Y;
                        a3 = x4 - 3 * x3 + 3 * x2 - x1;
                        a2 = 3 * x3 - 6 * x2 + 3 * x1;
                        a1 = 3 * (x2 - x1);
                        a0 = x1;
                        return ((a3 * t + a2) * t + a1) * t + a0;
                    }
                }
            }
            return Default;
        }

        /// <summary>
        /// 3次方程式a3*x^3 + a2*x^2 + a1*x + a0 = ansの解をニュートン法を使って計算します。ただし、単調増加である必要がある。
        /// </summary>
        /// <param name="a3"></param>
        /// <param name="a2"></param>
        /// <param name="a1"></param>
        /// <param name="a0"></param>
        /// <param name="ans"></param>
        /// <returns></returns>
        private static float SolveCubicEquation( float a3, float a2, float a1, float a0, float ans ) {
            const double EPSILON = 1e-9;
            double suggested_t = 0.5;
            double a3_3 = a3 * 3.0;
            double a2_2 = a2 * 2.0;
            while ( (a3_3 * suggested_t + a2_2) * suggested_t + a1 == 0.0 ) {
                suggested_t += 0.1;
            }
            double x = suggested_t;
            double new_x = suggested_t;
            for ( int i = 0; i < 5000; i++ ) {
                new_x = x - (((a3 * x + a2) * x + a1) * x + a0 - ans) / ((a3_3 * x + a2_2) * x + a1);
                if ( Math.Abs( new_x - x ) < EPSILON * new_x ) {
                    break;
                }
                x = new_x;
            }
            return (float)new_x;
        }
    }

}
