﻿/*
 * BezierPoint.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Drawing;
using System.Xml.Serialization;

namespace Boare.Cadencii {

    /// <summary>
    /// ベジエ曲線を構成するデータ点。
    /// </summary>
    [Serializable]
    public class BezierPoint : IComparable<BezierPoint>, ICloneable {
        Point m_base;
        internal Point m_control_left;
        internal Point m_control_right;
        BezierControlType m_type_left;
        BezierControlType m_type_right;
        [NonSerialized]
        int m_id;

        [XmlIgnore]
        public int ID {
            get {
                return m_id;
            }
            internal set {
                m_id = value;
            }
        }

        public override string ToString() {
            return "m_base=" + m_base.X + "," + m_base.Y + "\n" +
                "m_control_left=" + m_control_left.X + "," + m_control_left.Y + "\n" +
                "m_control_right=" + m_control_right.X + "," + m_control_right.Y + "\n" +
                "m_type_left=" + m_type_left + "\n" +
                "m_type_right=" + m_type_right + "\n";
        }

        public BezierPoint() {
        }

        public BezierPoint( Point p1, Point left, Point right ) {
            m_base = p1;
            m_control_left = new Point( left.X - m_base.X, left.Y - m_base.Y );
            m_control_right = new Point( right.X - m_base.X, right.Y - m_base.Y );
            m_type_left = BezierControlType.None;
            m_type_right = BezierControlType.None;
        }

        public object Clone() {
            BezierPoint result = new BezierPoint( this.Base, this.ControlLeft, this.ControlRight );
            result.m_control_left = this.m_control_left;
            result.m_control_right = this.m_control_right;
            result.m_type_left = this.m_type_left;
            result.m_type_right = this.m_type_right;
            result.m_id = this.m_id;
            return result;
        }

        public int CompareTo( BezierPoint item ) {
            if ( this.Base.X > item.Base.X ) {
                return 1;
            } else if ( this.Base.X < item.Base.X ) {
                return -1;
            } else {
                if ( this.ID > item.ID ) {
                    return 1;
                } else if ( this.ID < item.ID ) {
                    return -1;
                } else {
                    return 0;
                }
            }
        }

        public Point Base {
            get {
                return m_base;
            }
            set {
                m_base = value;
            }
        }

        public void SetPosition( BezierPickedSide picked_side, Point new_position ) {
            if ( picked_side == BezierPickedSide.Base ) {
                this.Base = new_position;
            } else if ( picked_side == BezierPickedSide.Left ) {
                this.m_control_left = new Point( new_position.X - this.Base.X, new_position.Y - this.Base.Y );
            } else {
                this.m_control_right = new Point( new_position.X - this.Base.X, new_position.Y - this.Base.Y );
            }
        }

        public PointF GetPosition( BezierPickedSide picked_side ) {
            if ( picked_side == BezierPickedSide.Base ) {
                return this.Base;
            } else if ( picked_side == BezierPickedSide.Left ) {
                return this.ControlLeft;
            } else {
                return this.ControlRight;
            }
        }

        public BezierControlType GetControlType( BezierPickedSide picked_side ) {
            if ( picked_side == BezierPickedSide.Left ) {
                return this.ControlLeftType;
            } else if ( picked_side == BezierPickedSide.Right ) {
                return this.ControlRightType;
            } else {
                return BezierControlType.None;
            }
        }

        public Point ControlLeft {
            get {
                if ( m_type_left != BezierControlType.None ) {
                    return new Point( m_base.X + m_control_left.X, m_base.Y + m_control_left.Y );
                } else {
                    return m_base;
                }
            }
            set {
                m_control_left = new Point( value.X - m_base.X, value.Y - m_base.Y );
            }
        }

        public Point ControlRight {
            get {
                if ( m_type_right != BezierControlType.None ) {
                    return new Point( m_base.X + m_control_right.X, m_base.Y + m_control_right.Y );
                } else {
                    return m_base;
                }
            }
            set {
                m_control_right = new Point( value.X - m_base.X, value.Y - m_base.Y );
            }
        }

        public BezierControlType ControlLeftType {
            get {
                return m_type_left;
            }
            set {
                m_type_left = value;
            }
        }

        public BezierControlType ControlRightType {
            get {
                return m_type_right;
            }
            set {
                m_type_right = value;
            }
        }
    }

}
