﻿/*
 * EditorConfig.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;
using System.IO;
using System.Windows.Forms;
using System.Drawing;

using bocoree;
using Boare.Lib.Vsq;

namespace Boare.Cadencii {

    /// <summary>
    /// Cadenciiの環境設定
    /// </summary>
    public class EditorConfig {
        public int DefaultPreMeasure = 4;
        public int DefaultSinger = 0;
        public int DefaultXScale = 65;
        public string BaseFontName = "MS UI Gothic";
        public string ScreenFontName = "MS UI Gothic";
        public string CounterFontName = "Arial";
        public int WheelOrder = 20;
        public bool CursorFixed = false;
        /// <summary>
        /// RecentFilesに登録することの出来る最大のファイル数
        /// </summary>
        public int NumRecentFiles = 5;
        /// <summary>
        /// 最近使用したファイルのリスト
        /// </summary>
        public List<string> RecentFiles = new List<string>();
        public int DefaultPMBendDepth = 8;
        public int DefaultPMBendLength = 0;
        public int DefaultPMbPortamentoUse = 3;
        public int DefaultDEMdecGainRate = 50;
        public int DefaultDEMaccent = 50;
        public bool ShowLyric = true;
        public bool ShowExpLine = true;
        public DefaultVibratoLength DefaultVibratoLength = DefaultVibratoLength.L66;
        public AutoVibratoMinLength AutoVibratoMinimumLength = AutoVibratoMinLength.L1;
        public VibratoType AutoVibratoType = VibratoType.NormalType1;
        public bool EnableAutoVibrato = true;
        public int PxTrackHeight = 14;
        //public string DllPath = @"C:\Program Files\Steinberg\VSTplugins\VOCALOID2\VOCALOID2.dll";
        //public bool AutoScroll = true;
        public int MouseDragIncrement = 50;
        public int MouseDragMaximumRate = 600;
        public bool MixerVisible = false;
        public int PreSendTime = 500;
        public ClockResolution ControlCurveResolution = ClockResolution.L30;
        public string Language = "ja";
        //public string ExpDbDir = @"C:\Program Files\VOCALOID2\voicedbdir";
        /// <summary>
        /// マウスの操作などの許容範囲。プリメジャーにPxToleranceピクセルめり込んだ入力を行っても、エラーにならない。(補正はされる)
        /// </summary>
        public int PxTolerance = 10;
        /// <summary>
        /// マウスホイールでピアノロールを水平方向にスクロールするかどうか。
        /// </summary>
        public bool ScrollHorizontalOnWheel = false;
        /// <summary>
        /// テンポ/拍子の変更ダイアログを、使いやすい位置に出すかどうか
        /// </summary>
        public bool CorrectPlacementOfTempoAndBeatDialog = false;
        /// <summary>
        /// 画面描画の最大フレームレート
        /// </summary>
        public int MaximumFrameRate = 30;
        /// <summary>
        /// ユーザー辞書のOn/Offと順序
        /// </summary>
        public List<string> UserDictionaries = new List<string>();
        /// <summary>
        /// 実行環境
        /// </summary>
        public Platform Platform = Platform.Windows;
        /// <summary>
        /// toolStripEditの表示位置
        /// </summary>
        public ToolStripLocation ToolEditLocation = new ToolStripLocation( new Point( 3, 0 ), ToolStripLocation.ParentPanel.Top );
        /// <summary>
        /// toolStripPositionの表示位置
        /// </summary>
        public ToolStripLocation ToolPositionLocation = new ToolStripLocation( new Point( 3, 25 ), ToolStripLocation.ParentPanel.Top );
        /// <summary>
        /// toolStripMeasureの表示位置
        /// </summary>
        public ToolStripLocation ToolMeasureLocation = new ToolStripLocation( new Point( 3, 50 ), ToolStripLocation.ParentPanel.Top );
        /// <summary>
        /// toolStripCursorの表示位置
        /// </summary>
        public ToolStripLocation ToolCursorLocation = new ToolStripLocation( new Point( 3, 75 ), ToolStripLocation.ParentPanel.Top );
        public bool WindowMaximized = false;
        public Rectangle WindowRect = new Rectangle( 0, 0, 756, 616 );
        /// <summary>
        /// hScrollのスクロールボックスの最小幅(px)
        /// </summary>
        public int MinimumScrollHandleWidth = 20;
        /// <summary>
        /// 発音記号入力モードを，維持するかどうか
        /// </summary>
        public bool KeepLyricInputMode = false;
        /// <summary>
        /// 最後に使用したVSQファイルへのパス
        /// </summary>
        public string LastVsqPath = "";
        /// <summary>
        /// ピアノロールの何もないところをクリックした場合、右クリックでもプレビュー音を再生するかどうか
        /// </summary>
        public bool PlayPreviewWhenRightClick = false;
        /// <summary>
        /// 鍵盤を押した時に鳴らすWAVEを作成する時使用するプログラムチェンジ
        /// </summary>
        public byte PreviewSingerProgramChange = 0x00;

        private QuantizeMode m_position_quantize = QuantizeMode.p32;
        private bool m_position_quantize_triplet = false;
        private QuantizeMode m_length_quantize = QuantizeMode.p32;
        private bool m_length_quantize_triplet = false;
        private int m_mouse_hover_time = 500;
        
        private static XmlSerializer s_serizlizer = null;

        /// <summary>
        /// PositionQuantize, PositionQuantizeTriplet, LengthQuantize, LengthQuantizeTripletの描くプロパティのいずれかが
        /// 変更された時発生します
        /// </summary>
        public static event EventHandler QuantizeModeChanged;

        /// <summary>
        /// ピアノロール上でマウスホバーイベントが発生するまでの時間(millisec)
        /// </summary>
        public int MouseHoverTime {
            get {
                return m_mouse_hover_time;
            }
            set {
                if ( value < 0 ) {
                    m_mouse_hover_time = 0;
                } else if ( 2000 < m_mouse_hover_time ) {
                    m_mouse_hover_time = 2000;
                } else {
                    m_mouse_hover_time = value;
                }
            }
        }

        public QuantizeMode PositionQuantize {
            get {
                return m_position_quantize;
            }
            set {
                if ( m_position_quantize != value ) {
                    m_position_quantize = value;
                    if ( QuantizeModeChanged != null ) {
                        QuantizeModeChanged( typeof( EditorConfig ), new EventArgs() );
                    }
                }
            }
        }

        public bool PositionQuantizeTriplet {
            get {
                return m_position_quantize_triplet;
            }
            set {
                if ( m_position_quantize_triplet != value ) {
                    m_position_quantize_triplet = value;
                    if ( QuantizeModeChanged != null ) {
                        QuantizeModeChanged( typeof( EditorConfig ), new EventArgs() );
                    }
                }
            }
        }

        public QuantizeMode LengthQuantize {
            get {
                return m_length_quantize;
            }
            set {
                if ( m_length_quantize != value ) {
                    m_length_quantize = value;
                    if ( QuantizeModeChanged != null ) {
                        QuantizeModeChanged( typeof( EditorConfig ), new EventArgs() );
                    }
                }
            }
        }

        public bool LengthQuantizeTriplet {
            get {
                return m_length_quantize_triplet;
            }
            set {
                if ( m_length_quantize_triplet != value ) {
                    m_length_quantize_triplet = value;
                    if ( QuantizeModeChanged != null ) {
                        QuantizeModeChanged( typeof( EditorConfig ), new EventArgs() );
                    }
                }
            }
        }

        public void Save( string file ) {
            // ユーザー辞書の情報を取り込む
            UserDictionaries.Clear();
            for ( int i = 0; i < SymbolTable.Count; i++ ) {
                UserDictionaries.Add( SymbolTable.GetSymbolTable( i ).Name + "\t" + (SymbolTable.GetSymbolTable( i ).Enabled ? "T" : "F") );
            }

            if ( s_serizlizer == null ) {
                s_serizlizer = new XmlSerializer( typeof( EditorConfig ) );
            }
            using ( FileStream fs = new FileStream( file, FileMode.Create ) ) {
                s_serizlizer.Serialize( fs, this );
            }
        }
        
        /// <summary>
        /// 「最近使用したファイル」のリストに、アイテムを追加します
        /// </summary>
        /// <param name="new_file"></param>
        public void PushRecentFiles( string new_file ) {
#if DEBUG
            System.Diagnostics.Debug.WriteLine( "PushRecentFiles" );
#endif
            // NumRecentFilesは0以下かも知れない
            if ( NumRecentFiles <= 0 ) {
                NumRecentFiles = 5;
            }

            // RecentFilesはnullかもしれない．
            if ( RecentFiles == null ) {
                RecentFiles = new List<string>();
            }

            // 重複があれば消す
            List<string> dict = new List<string>();
            foreach ( string s in RecentFiles ) {
                bool found = false;
                for ( int i = 0; i < dict.Count; i++ ) {
                    if ( s == dict[i] ) {
                        found = true;
                    }
                }
                if ( !found ) {
                    dict.Add( s );
                }
            }
            RecentFiles.Clear();
            foreach ( string s in dict ) {
                RecentFiles.Add( s );
            }

            // 現在登録されているRecentFilesのサイズが規定より大きければ，下の方から消す
            if ( RecentFiles.Count > NumRecentFiles ) {
                for ( int i = RecentFiles.Count - 1; i > NumRecentFiles; i-- ) {
                    RecentFiles.RemoveAt( i );
                }
            }

            // 登録しようとしているファイルは，RecentFilesの中に既に登録されているかs？
            int index = -1;
            for ( int i = 0; i < RecentFiles.Count; i++ ) {
                if ( RecentFiles[i] == new_file ) {
                    index = i;
                    break;
                }
            }

            if ( index >= 0 ) {  // 登録されてる場合
                RecentFiles.RemoveAt( index );
            }
            RecentFiles.Insert( 0, new_file );
        }
    }

}
