﻿/*
 * EditorManager.cs
 * Copyright (c) 2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using System.Threading;

using Boare.Lib.Vsq;

namespace Boare.Cadencii {

    class EditorManager {
        private int m_start_to_draw_y = 0;
        private VsqFileEx m_vsq;
        private string m_file = "";
        private int m_selected = 1;
        private int m_current_clock = 0;
        private bool m_playing = false;
        private bool m_repeat_mode = false;
        private bool m_saveavi = false;
        private bool m_grid_visible = false;
        private EditMode m_edit_mode = EditMode.None;
        /// <summary>
        /// トラックのON/OFFを表す
        /// </summary>
        private bool[] m_is_on = new bool[16] { true, true, true, true, true, true, true, true, true, true, true, true, true, true, true, true };
        /// <summary>
        /// トラックのレンダリングが必要かどうかを表すフラグ
        /// </summary>
        private bool[] m_render_required = new bool[16] { false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false };
        /// <summary>
        /// トラックのオーバーレイ表示
        /// </summary>
        private bool m_overlay = true;
        private Dictionary<int, SelectedEventEntry> m_selected_event = new Dictionary<int, SelectedEventEntry>();
        /// <summary>
        /// 選択されているテンポ変更イベントのリスト
        /// </summary>
        private Dictionary<int, SelectedTempoEntry> m_selected_tempo = new Dictionary<int, SelectedTempoEntry>();
        private int m_last_selected_tempo = -1;
        /// <summary>
        /// 選択されている拍子変更イベントのリスト
        /// </summary>
        private Dictionary<int, SelectedTimesigEntry> m_selected_timesig = new Dictionary<int, SelectedTimesigEntry>();
        private int m_last_selected_timesig = -1;
        private EditTool m_selected_tool = EditTool.Arrow;
        private List<CadenciiCommand> m_commands = new List<CadenciiCommand>();
        private int m_command_position = -1;
        /// <summary>
        /// 選択されているベジエ点のリスト
        /// </summary>
        private List<SelectedBezierPoint> m_selected_bezier = new List<SelectedBezierPoint>();
        /// <summary>
        /// 最後に選択されたベジエ点
        /// </summary>
        private SelectedBezierPoint m_last_selected_bezier = new SelectedBezierPoint();
        /// <summary>
        /// 現在の演奏カーソルの位置(m_current_clockと意味は同じ。CurrentClockが変更されると、自動で更新される)
        /// </summary>
        private PlayPositionSpecifier m_current_play_position;

        public bool SelectedRegionEnabled = false;
        /// <summary>
        /// Ctrlキーを押しながらのマウスドラッグ操作による選択が行われた範囲
        /// </summary>
        public SelectedRegion SelectedRegion;
        /// <summary>
        /// 自動ノーマライズモードかどうかを表す値を取得、または設定します。
        /// </summary> 
        public bool AutoNormalize = false;
        /// <summary>
        /// エンドマーカーの位置(clock)
        /// </summary>
        public int EndMarker = 0;
        public bool EndMarkerEnabled = false;
        /// <summary>
        /// x方向の表示倍率(pixel/clock)
        /// </summary>
        public float ScaleX = 0.1f;
        /// <summary>
        /// プレビューが開始された時刻
        /// </summary>
        public DateTime Started;
        /// <summary>
        /// スタートマーカーの位置(clock)
        /// </summary>
        public int StartMarker = 0;
        /// <summary>
        /// Bezierカーブ編集モードが有効かどうかを表す
        /// </summary>
        public bool m_is_curve_mode = false;
        public bool StartMarkerEnabled = false;
        public AppManager FormRoot;
        /// <summary>
        /// FormMainを識別するID
        /// </summary>
        public string ID = "";
        /// <summary>
        /// 再生時に自動スクロールするかどうか
        /// </summary>
        public bool AutoScroll = true;

        /// <summary>
        /// グリッド線を表示するか否かを表す値が変更された時発生します
        /// </summary>
        public event EventHandler GridVisibleChanged;
        public event EventHandler PreviewStarted;
        public event EventHandler PreviewAborted;
        public event BSimpleDelegate<bool> SelectedEventChanged;
        public event EventHandler SelectedToolChanged;
        /// <summary>
        /// CurrentClockプロパティの値が変更された時発生します
        /// </summary>
        public event EventHandler CurrentClockChanged;
        /// <summary>
        /// このインスタンスが担当するFormMainが閉じられた時発生します
        /// </summary>
        public event AppManager.MainFormClosedEventHandler MainFormClosed;
        /// <summary>
        /// このインスタンスが担当するFormMainにおいてEditorConfigの内容が変更されたとき発生します
        /// </summary>
        public event AppManager.EditorConfigChangedEventHandler EditorConfigChanged;
        /// <summary>
        /// このインスタンスが担当するFormMainの表示を，EditorConfigの内容に応じて変更することが求められたとき発生します
        /// </summary>
        public event EventHandler ApplyEditorConfigRequired;

        private const string _TEMPDIR_NAME = "temp";

        public EditorManager(){
        }

        public void SetFormRoot( AppManager root ) {
            FormRoot = root;
        }

        public string TempWaveDir {
            get {
                return Path.Combine( Path.Combine( Application.StartupPath, _TEMPDIR_NAME ), ID );
            }
        }

        /// <summary>
        /// (FormMain宛てに)EditorConfigの内容に応じて表示内容を変更することを要求します
        /// </summary>
        public void SendApplyEditorConfigMessage() {
            if ( ApplyEditorConfigRequired != null ) {
                ApplyEditorConfigRequired( this, null );
            }
        }

        /// <summary>
        /// (FormRoot宛てに)EditorConfigが変更されたことを知らせます
        /// </summary>
        public void SendEditorConfigChangedMessage() {
            if ( EditorConfigChanged != null ) {
                EditorConfigChanged( this );
            }
        }

        /// <summary>
        /// (FormRoot宛てに)FormMainが閉じられたことを知らせます
        /// </summary>
        public void SendWindowClosedMessage() {
            if ( MainFormClosed != null ) {
                MainFormClosed( this );
            }
        }

        public bool IsCurveMode {
            get {
                return m_is_curve_mode;
            }
            set {
                bool old = m_is_curve_mode;
                m_is_curve_mode = value;
                if ( old != m_is_curve_mode && SelectedToolChanged != null ) {
                    SelectedToolChanged( typeof( AppManager ), new EventArgs() );
                }
            }
        }

        /// <summary>
        /// コマンドの履歴を削除します
        /// </summary>
        public void ClearCommandBuffer() {
            m_commands.Clear();
            m_command_position = -1;
        }

        /// <summary>
        /// アンドゥ処理を行います
        /// </summary>
        public void Undo() {
#if DEBUG
            Common.DebugWriteLine( "CommonConfig.Undo()" );
#endif
            if ( IsUndoAvailable ) {
                CadenciiCommand run = m_commands[m_command_position];
#if DEBUG
                Common.DebugWriteLine( "    command type=" + run.Type );
#endif
                if ( run.VsqCommand != null ) {
                    if ( run.VsqCommand.Type == VsqCommandType.DeleteTrack ) {
                        int track = (int)run.Args[0];
                        if ( track == Selected && track >= 2 ) {
                            Selected = track - 1;
                        }
                    }
                }
                CadenciiCommand inv = m_vsq.Execute( run );
                m_commands[m_command_position] = inv;
                m_command_position--;
            }
        }

        /// <summary>
        /// リドゥ処理を行います
        /// </summary>
        public void Redo() {
#if DEBUG
            Common.DebugWriteLine( "CommonConfig.Redo()" );
#endif
            if ( IsRedoAvailable ) {
                CadenciiCommand run = m_commands[m_command_position + 1];
                if ( run.VsqCommand != null ) {
                    if ( run.VsqCommand.Type == VsqCommandType.DeleteTrack ) {
                        int track = (int)run.Args[0];
                        if ( track == Selected && track >= 2 ) {
                            Selected = track - 1;
                        }
                    }
                }
                CadenciiCommand inv = m_vsq.Execute( run );
                m_commands[m_command_position + 1] = inv;
                m_command_position++;
            }
        }

        /// <summary>
        /// コマンドバッファに指定されたコマンドを登録します
        /// </summary>
        /// <param name="command"></param>
        public void Register( CadenciiCommand command ) {
            if ( m_command_position == m_commands.Count - 1 ) {
                // 新しいコマンドバッファを追加する場合
                m_commands.Add( command );
                m_command_position = m_commands.Count - 1;
            } else {
                // 既にあるコマンドバッファを上書きする場合
                m_commands[m_command_position + 1] = command;
                for ( int i = m_commands.Count - 1; i >= m_command_position + 2; i-- ) {
                    m_commands.RemoveAt( i );
                }
                m_command_position++;
            }
        }

        /// <summary>
        /// リドゥ操作が可能かどうかを表すプロパティ
        /// </summary>
        public bool IsRedoAvailable {
            get {
                if ( m_commands.Count > 0 && 0 <= m_command_position + 1 && m_command_position + 1 < m_commands.Count ) {
                    return true;
                } else {
                    return false;
                }
            }
        }

        /// <summary>
        /// アンドゥ操作が可能かどうかを表すプロパティ
        /// </summary>
        public bool IsUndoAvailable {
            get {
                if ( m_commands.Count > 0 && 0 <= m_command_position && m_command_position < m_commands.Count ) {
                    return true;
                } else {
                    return false;
                }
            }
        }

        public EditTool SelectedTool {
            get {
                return m_selected_tool;
            }
            internal set {
                EditTool old = m_selected_tool;
                m_selected_tool = value;
                if ( old != m_selected_tool && SelectedToolChanged != null ) {
                    SelectedToolChanged( typeof( AppManager ), new EventArgs() );
                }
            }
        }

        #region SelectedBezier
        public IEnumerable<SelectedBezierPoint> GetSelectedBezierEnumerator() {
            for ( int i = 0; i < m_selected_bezier.Count; i++ ) {
                yield return m_selected_bezier[i];
            }
        }

        /// <summary>
        /// ベジエ点のリストに最後に追加された点の情報を取得します
        /// </summary>
        public SelectedBezierPoint LastSelectedBezier {
            get {
                if ( m_last_selected_bezier.ChainID < 0 || m_last_selected_bezier.PointID < 0 ) {
                    return null;
                } else {
                    return m_last_selected_bezier;
                }
            }
        }

        /// <summary>
        /// 選択されているベジエ点のリストに、アイテムを追加します
        /// </summary>
        /// <param name="selected">追加する点</param>
        public void AddSelectedBezier( SelectedBezierPoint selected ) {
            m_last_selected_bezier = selected;
            int index = -1;
            for ( int i = 0; i < m_selected_bezier.Count; i++ ) {
                if ( m_selected_bezier[i].ChainID == selected.ChainID &&
                    m_selected_bezier[i].PointID == selected.PointID ) {
                    index = i;
                    break;
                }
            }
            if ( index >= 0 ) {
                m_selected_bezier[index] = selected;
            } else {
                m_selected_bezier.Add( selected );
            }
        }

        /// <summary>
        /// 選択されているベジエ点のリストを初期化します
        /// </summary>
        public void ClearSelectedBezier() {
            m_selected_bezier.Clear();
            m_last_selected_bezier.ChainID = -1;
            m_last_selected_bezier.PointID = -1;
        }
        #endregion

        #region SelectedTimesig
        /// <summary>
        /// 選択されている拍子変更イベントのリスト
        /// </summary>
        public Dictionary<int, SelectedTimesigEntry> SelectedTimesig {
            get {
                return m_selected_timesig;
            }
        }

        public SelectedTimesigEntry LastSelectedTimesig {
            get {
                if ( m_selected_timesig.ContainsKey( m_last_selected_timesig ) ) {
                    return m_selected_timesig[m_last_selected_timesig];
                } else {
                    return null;
                }
            }
        }

        public int LastSelectedTimesigBarcount {
            get {
                return m_last_selected_timesig;
            }
        }

        public void AddSelectedTimesig( int barcount ) {
            ClearSelectedEvent(); //ここ注意！
            ClearSelectedTempo();
            m_last_selected_timesig = barcount;
            if ( !m_selected_timesig.ContainsKey( barcount ) ) {
                foreach ( TimeSigTableEntry tte in m_vsq.TimeSigTable ) {
                    if ( tte.BarCount == barcount ) {
                        m_selected_timesig.Add( barcount, new SelectedTimesigEntry( tte, (TimeSigTableEntry)tte.Clone() ) );
                        break;
                    }
                }
            }
        }

        public void ClearSelectedTimesig() {
            m_selected_timesig.Clear();
            m_last_selected_timesig = -1;
        }
        #endregion

        #region SelectedTempo
        /// <summary>
        /// 選択されているテンポ変更イベントのリスト
        /// </summary>
        public Dictionary<int, SelectedTempoEntry> SelectedTempo {
            get {
                return m_selected_tempo;
            }
        }

        public SelectedTempoEntry LastSelectedTempo {
            get {
                if ( m_selected_tempo.ContainsKey( m_last_selected_tempo ) ) {
                    return m_selected_tempo[m_last_selected_tempo];
                } else {
                    return null;
                }
            }
        }

        public int LastSelectedTempoClock {
            get {
                return m_last_selected_tempo;
            }
        }

        public void AddSelectedTempo( int clock ) {
            ClearSelectedEvent(); //ここ注意！
            ClearSelectedTimesig();
            m_last_selected_tempo = clock;
            if ( !m_selected_tempo.ContainsKey( clock ) ) {
                foreach ( TempoTableEntry tte in m_vsq.TempoTable ) {
                    if ( tte.Clock == clock ) {
                        m_selected_tempo.Add( clock, new SelectedTempoEntry( tte, (TempoTableEntry)tte.Clone() ) );
                        break;
                    }
                }
            }
        }

        public void ClearSelectedTempo() {
            m_selected_tempo.Clear();
            m_last_selected_tempo = -1;
        }
        #endregion

        #region SelectedEvent
        public Dictionary<int, SelectedEventEntry> SelectedEvent {
            get {
                return m_selected_event;
            }
        }

        public SelectedEventEntry LastSelectedEvent {
            get {
                int id_last = LastSelectedEventID;
                if ( m_selected_event.ContainsKey( id_last ) ) {
                    return m_selected_event[id_last];
                } else {
                    return null;
                }
            }
        }

        public int LastSelectedEventID {
            get {
                int max = -1;
                int max_id = -1;
                foreach ( KeyValuePair<int, SelectedEventEntry> item in m_selected_event ) {
                    if ( max < item.Value.Order ) {
                        max = item.Value.Order;
                        max_id = item.Key;
                    }
                }
                return max_id;
            }
        }

        public void AddSelectedEvent( int id ) {
            ClearSelectedTempo();
            ClearSelectedTimesig();
            if ( !m_selected_event.ContainsKey( id ) ) {
                int max = 0;
                foreach ( SelectedEventEntry see in m_selected_event.Values ) {
                    max = Math.Max( max, see.Order );
                }
                max++;
                foreach ( VsqEvent ev in m_vsq.Tracks[m_selected].Events ) {
                    if ( ev.InternalID == id ) {
                        m_selected_event.Add( id, new SelectedEventEntry( ev, (VsqEvent)ev.Clone(), max ) );
                        if ( SelectedEventChanged != null ) {
                            SelectedEventChanged( false );
                        }
                        break;
                    }
                }
            } else {
                int max = 0;
                foreach ( SelectedEventEntry see in m_selected_event.Values ) {
                    max = Math.Max( max, see.Order );
                }
                max++;
                m_selected_event[id].Order = max;
            }
        }

        public void ClearSelectedEvent() {
            m_selected_event.Clear();
            if ( SelectedEventChanged != null ) {
                SelectedEventChanged( true );
            }
        }
        #endregion

        public bool Overlay {
            get {
                return m_overlay;
            }
            set {
                m_overlay = value;
            }
        }

        public bool getRenderRequired( int track ) {
            return m_render_required[track - 1];
        }

        public void setRenderRequired( int track, bool value ) {
            m_render_required[track - 1] = value;
        }

        public bool getTrackEnabled( int track ) {
            return m_is_on[track - 1];
        }

        public void setTrackEnabled( int track, bool value ) {
            m_is_on[track - 1] = value;
        }

        /// <summary>
        /// 現在の編集モード
        /// </summary>
        public EditMode EditMode {
            get {
                return m_edit_mode;
            }
            set {
                m_edit_mode = value;
            }
        }

        /// <summary>
        /// グリッドを表示するか否かを表すフラグを取得または設定します
        /// </summary>
        public bool GridVisible {
            get {
                return m_grid_visible;
            }
            set {
                if ( value != m_grid_visible ) {
                    m_grid_visible = value;
                    if ( GridVisibleChanged != null ) {
                        GridVisibleChanged( typeof( AppManager ), new EventArgs() );
                    }
                }
            }
        }

        /// <summary>
        /// AVIキャプチャ中かどうかを表す値を取得します
        /// </summary>
        public bool AviSaving {
            get {
                return m_saveavi;
            }
            set {
                m_saveavi = value;
            }
        }

        /// <summary>
        /// 現在のプレビューがリピートモードであるかどうかを表す値を取得または設定します
        /// </summary>
        public bool IsRepeatMode {
            get {
                return m_repeat_mode;
            }
            set {
                m_repeat_mode = value;
            }
        }

        /// <summary>
        /// 現在プレビュー中かどうかを示す値を取得または設定します
        /// </summary>
        public bool Playing {
            get {
                return m_playing;
            }
            set {
                bool previous = m_playing;
                m_playing = value;
                if ( previous != m_playing ) {
                    if ( m_playing && PreviewStarted != null ) {
                        int clock = CurrentClock;
                        PreviewStarted( this, new EventArgs() );
                    } else if ( !m_playing && PreviewAborted != null ) {
                        PreviewAborted( this, new EventArgs() );
                    }
                }
            }
        }

        /// <summary>
        /// _vsq_fileにセットされたvsqファイルの名前を取得します。
        /// </summary>
        public string FileName {
            get {
                return m_file;
            }
        }

        public void SaveTo( string file ) {
            if ( m_vsq != null ) {
                string path = Path.GetDirectoryName( file );
                bool save_separate = false;
                string ext = Path.GetExtension( file ).ToLower();
                if ( ext == ".vsq" || ext == ".mid" ) {
                    save_separate = true;
                } else {
                    save_separate = false;
                }
                if ( save_separate ) {
                    m_vsq.Write( file );
                    m_file = file;
                    List<BezierCurves> tmp = new List<BezierCurves>();
                    for ( int i = 0; i < m_vsq.AttachedCurves.Count; i++ ) {
                        tmp.Add( new BezierCurves( m_vsq.AttachedCurves[i] ) );
                    }
                    using ( FileStream fs = new FileStream( Path.Combine( path, file + ".xml" ), FileMode.Create ) ) {
                        AppManager.XmlSerializerListBezierCurves.Serialize( fs, tmp );
                    }
                } else {
                    BinaryFormatter bf = new BinaryFormatter();
                    file = Path.Combine( path, Path.GetFileNameWithoutExtension( file ) + ".evsq" );
                    using ( FileStream fs = new FileStream( file, FileMode.Create ) ) {
                        bf.Serialize( fs, m_vsq );
                    }
                    m_file = file;
                }
            }
        }

        /// <summary>
        /// 位置clockにおけるテンポ、拍子、小節数を取得します
        /// </summary>
        /// <param name="clock"></param>
        /// <param name="numerator"></param>
        /// <param name="denominator"></param>
        /// <param name="bar_count"></param>
        /// <param name="tempo"></param>
        public void GetTempoAndTimeSignature(
            int clock,
            out int numerator,
            out int denominator,
            out int bar_count,
            out int tempo,
            out int time_sig_index,
            out int tempo_index,
            out int last_clock ) {
            numerator = 4;
            denominator = 4;
            bar_count = 0;
            tempo = 480000;
            time_sig_index = -1;
            tempo_index = -1;
            last_clock = 0;
            if ( m_vsq != null ) {
                int index = -1;
                for ( int i = 0; i < m_vsq.TimeSigTable.Count; i++ ) {
                    if ( m_vsq.TimeSigTable[i].Clock > clock ) {
                        index = i - 1;
                        break;
                    }
                }
                if ( index >= 0 ) {
                    denominator = m_vsq.TimeSigTable[index].Denominator;
                    numerator = m_vsq.TimeSigTable[index].Numerator;
                    time_sig_index = index;
                    last_clock = m_vsq.TimeSigTable[index].Clock;
                }
                bar_count = m_vsq.GetBarCountFromClock( clock );

                index = -1;
                for ( int i = 0; i < m_vsq.TempoTable.Count; i++ ) {
                    if ( m_vsq.TempoTable[i].Clock > clock ) {
                        index = i - 1;
                        break;
                    }
                }
                if ( index >= 0 ) {
                    tempo = m_vsq.TempoTable[index].Tempo;
                    tempo_index = index;
                }
            }
        }

        /// <summary>
        /// 現在の演奏マーカーの位置を取得または設定します。
        /// </summary>
        public int CurrentClock {
            get {
                return m_current_clock;
            }
            set {
                int old = m_current_clock;
                m_current_clock = value;
                int barcount = m_vsq.GetBarCountFromClock( m_current_clock );
                int bar_top_clock = m_vsq.GetClockFromBarCount( barcount );
                int numerator = 4;
                int denominator = 4;
                m_vsq.GetTimesigAt( m_current_clock, out numerator, out denominator );
                int clock_per_beat = 480 / 4 * denominator;
                int beat = (m_current_clock - bar_top_clock) / clock_per_beat;
                m_current_play_position.BarCount = barcount - m_vsq.PreMeasure + 1;
                m_current_play_position.Beat = beat + 1;
                m_current_play_position.Clock = m_current_clock - bar_top_clock - clock_per_beat * beat;
                m_current_play_position.Denominator = denominator;
                m_current_play_position.Numerator = numerator;
                m_current_play_position.Tempo = m_vsq.GetTempoAt( m_current_clock );
                if ( old != m_current_clock && CurrentClockChanged != null ) {
                    CurrentClockChanged( typeof( AppManager ), new EventArgs() );
                }
            }
        }

        /// <summary>
        /// 現在の演奏カーソルの位置(m_current_clockと意味は同じ。CurrentClockが変更されると、自動で更新される)
        /// </summary>
        public PlayPositionSpecifier PlayPosition {
            get {
                return m_current_play_position;
            }
        }

        /// <summary>
        /// 現在選択されているトラックを取得または設定します
        /// </summary>
        public int Selected {
            get {
                return m_selected;
            }
            set {
                m_selected = value;
            }
        }

        /// <summary>
        /// vsqファイルを読込みます
        /// </summary>
        /// <param name="file"></param>
        public void ReadVsq( string file ) {
            m_file = file;
            string ext = Path.GetExtension( file ).ToLower();
            if ( ext == ".vsq" || ext == ".mid" ) {
                if ( m_vsq != null ) {
                    m_vsq.Dispose();
                    m_vsq = null;
                }
                m_vsq = new VsqFileEx( file );
                string path = Path.GetDirectoryName( file );
                string xml = Path.Combine( path, Path.GetFileName( file ) + ".xml" );
                if ( File.Exists( xml ) ) {
                    List<BezierCurves> tmp = null;
                    FileStream fs = null;
                    try {
                        fs = new FileStream( xml, FileMode.Open );
                        tmp = (List<BezierCurves>)AppManager.XmlSerializerListBezierCurves.Deserialize( fs );
                    } catch {

                    } finally {
                        if ( fs != null ) {
                            fs.Close();
                        }
                    }
                    if ( tmp != null ) {
                        if ( tmp.Count == m_vsq.Tracks.Count - 1 ) {
                            m_vsq.AttachedCurves = new List<AttachedCurve>();
                            for ( int i = 0; i < tmp.Count; i++ ) {
                                m_vsq.AttachedCurves.Add( new AttachedCurve( tmp[i] ) );
                            }
                        }
                    }
                }
            } else {
                try {
                    BinaryFormatter bf = new BinaryFormatter();
                    using ( FileStream fs = new FileStream( file, FileMode.Open ) ) {
                        m_vsq = (VsqFileEx)bf.Deserialize( fs );
                    }
                } catch {

                }
            }
            StartMarker = m_vsq.PreMeasureClocks;
            int bar = m_vsq.PreMeasure + 1;
            EndMarker = m_vsq.GetClockFromBarCount( bar );
            if ( m_vsq.Tracks.Count >= 1 ) {
                m_selected = 1;
            } else {
                m_selected = -1;
            }
        }

        /// <summary>
        /// vsqファイル。
        /// </summary>
        public VsqFileEx VsqFile {
            get {
                return m_vsq;
            }
        }

        public void SetVsqFile( VsqFileEx vsq, string file_name ) {
            m_vsq = vsq;
            m_file = file_name;
            StartMarker = m_vsq.PreMeasureClocks;
            int bar = m_vsq.PreMeasure + 1;
            EndMarker = m_vsq.GetClockFromBarCount( bar );
        }

        /// <summary>
        /// 現在表示されているピアノロール画面の右上の、仮想スクリーン上座標で見たときのy座標(pixel)
        /// </summary>
        public int StartToDrawY {
            get {
                return m_start_to_draw_y;
            }
            set {
                m_start_to_draw_y = value;
            }
        }
    }

}
