﻿/*
 * FormScriptEditor.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.IO;
using System.Windows.Forms;

using Boare.Lib.AppUtil;

namespace Boare.Cadencii {

    partial class FormScriptEditor : Form {
        private string m_file = "";
        private bool m_edited = false;
        private EditorManager m_manager = null;
        
        public FormScriptEditor( EditorManager manager ) {
            InitializeComponent();
            m_manager = manager;
        }

        private void UpdateFormTitle() {
            string title = "";
            if ( m_file.Length > 0 ) {
                title = _( "Script Editor" ) + " : " + m_file;
            } else {
                title = _( "Script Editor" ) + " : Untitled";
            }
            if ( m_edited ) {
                title += " *";
            }
            Text = title;
        }

        public void ApplyLanguage() {
            UpdateFormTitle();
            menuFile.Text = _( "File" ) + "(&F)";
            menuFileOpen.Text = _( "Open" ) + "(&O)";
            menuFileSave.Text = _( "Save" ) + "(&S)";
            menuFileSaveAs.Text = _( "Save As" ) + "(&A)";
            menuFileQuit.Text = _( "Quit" ) + "(&Q)";

            menuBuild.Text = _( "Build" ) + "(&B)";
            menuBuildCompile.Text = _( "Compile" ) + "(&C)";
            menuBuildRun.Text = _( "Run" ) + "(&R)";

            menuSetting.Text = _( "Setting" ) + "(&S)";
            menuSettingFont.Text = _( "Font" ) + "(&F)";
        }

        public static string _( string id ) {
            return Messaging.GetMessage( id );
        }

        private void menuBuildCompile_Click( object sender, EventArgs e ) {
            string script = txtCode.Text;
            string err_msg;
            txtResult.Text = "Compile...\r\n";
            EditVsqScriptDelegate evsd = Common.LoadScript( script, AppManager.EditorConfig, out err_msg );
            err_msg = err_msg.Replace( "\n", "\r\n" );
            txtResult.Text += err_msg + "\r\n";
            if ( evsd != null ) {
                txtResult.Text += "Successed\r\n";
                evsd = null;
            } else {
                txtResult.Text += "Failed\r\n";
            }
        }

        private void FormScriptEditor_FormClosing( object sender, FormClosingEventArgs e ) {
            e.Cancel = true;
            if ( m_edited ) {
                string name = (m_file.Length > 0) ? m_file : "Untitled";
                DialogResult dr = MessageBox.Show(
                    string.Format( _( "{0}\nSave before closing?" ), name ),
                    _( "Affirmation" ),
                     MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question );
                if ( dr == DialogResult.Yes ) {
                    if ( m_file.Length <= 0 ) {
                        using ( SaveFileDialog sfd = new SaveFileDialog() ) {
                            sfd.Filter = GetFileDialogFilter();
                            if ( sfd.ShowDialog() == DialogResult.OK ) {
                                m_file = sfd.FileName;
                                SaveTo( m_file );
                                m_edited = false;
                                Hide();
                            }
                        }
                    } else {
                        SaveTo( m_file );
                        m_edited = false;
                        Hide();
                    }
                } else if ( dr == DialogResult.No ) {
                    Hide();
                }
            } else {
                Hide();
            }
        }

        private void FormScriptEditor_VisibleChanged( object sender, EventArgs e ) {
            ApplyLanguage();
        }

        private void menuFileOpen_Click( object sender, EventArgs e ) {
            using ( OpenFileDialog ofd = new OpenFileDialog() ) {
                ofd.Filter = GetFileDialogFilter();
                if ( ofd.ShowDialog() == DialogResult.OK ) {
                    string file = ofd.FileName;
                    txtCode.Text = "";
                    using ( StreamReader sr = new StreamReader( file ) ) {
                        string lines = sr.ReadToEnd();
                        lines = lines.Replace( "\r\n", "\n" );
                        lines = lines.Replace( "\n", "\r\n" );
                        txtCode.Text = lines;
                    }
                    m_edited = false;
                    m_file = file;
                    UpdateFormTitle();
                }
            }
        }

        private void menuFileSave_Click( object sender, EventArgs e ) {
            if ( m_file.Length <= 0 ) {
                using ( SaveFileDialog sfd = new SaveFileDialog() ) {
                    sfd.Filter = GetFileDialogFilter();
                    if ( sfd.ShowDialog() == DialogResult.OK ) {
                        m_file = sfd.FileName;
                    } else {
                        return;
                    }
                }
            }
            SaveTo( m_file );
            m_edited = false;
            UpdateFormTitle();
        }

        private void SaveTo( string file ) {
            using ( StreamWriter sw = new StreamWriter( file ) ) {
                sw.Write( txtCode.Text );
            }
        }

        private static string GetFileDialogFilter() {
            return _( "Cadencii Script(*.ves)|*.ves" ) + "|" + _( "All Files(*.*)|*.*" );
        }

        private void menuFileSaveAs_Click( object sender, EventArgs e ) {
            using ( SaveFileDialog sfd = new SaveFileDialog() ) {
                sfd.Filter = GetFileDialogFilter();
                if ( sfd.ShowDialog() == DialogResult.OK ) {
                    m_file = sfd.FileName;
                } else {
                    return;
                }
            }
            SaveTo( m_file );
            m_edited = false;
            UpdateFormTitle();
        }

        private void txtCode_TextChanged( object sender, EventArgs e ) {
            if ( !m_edited ) {
                m_edited = true;
                UpdateFormTitle();
            }
        }

        private void menuFileQuit_Click( object sender, EventArgs e ) {
            Close();
        }

        private void menuSettingFont_Click( object sender, EventArgs e ) {
            using ( FontDialog fd = new FontDialog() ) {
                fd.Font = txtCode.Font;
                if ( fd.ShowDialog() == DialogResult.OK ) {
                    txtCode.Font = fd.Font;
                }
            }
        }

        private void menuBuildRun_Click( object sender, EventArgs e ) {
            string err_msg;
            EditVsqScriptDelegate evsd = Common.LoadScript( txtCode.Text, AppManager.EditorConfig, out err_msg );
            if ( evsd != null ) {
                try {
                    evsd.Invoke( m_manager.VsqFile );
                } catch ( Exception ex ) {
                    txtResult.Text += ex.StackTrace;
                    if ( ex.InnerException != null ) {
                        txtResult.Text += ex.InnerException.StackTrace;
                    }
                }
            }
        }

        private void FormScriptEditor_Load( object sender, EventArgs e ) {
            PlatformID platform = Environment.OSVersion.Platform;
            menuFileSave.ShortcutKeys = (AppManager.EditorConfig.Platform == Platform.Macintosh ? Keys.Menu : Keys.Control) | Keys.S;
        }
    }

}
