﻿/*
 * FormSynthesize.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Forms;

using Boare.Lib.AppUtil;
using Boare.Lib.Vsq;

namespace Boare.Cadencii {

    partial class FormSynthesize : Form {
        private VsqFile m_vsq;
        private int m_presend = 500;
        private int[] m_tracks;
        private string[] m_files;
        private int m_start;
        private int m_end;
        private bool m_partial_mode = false;
        private int m_temp_premeasure = 0;
        private int m_finished = -1;
        private EditorManager m_manager = null;
        private bool m_rendering_started = false;

        public FormSynthesize( EditorManager manager, VsqFile vsq, int presend, int track, string file, int start, int end, int temp_premeasure ) {
            m_manager = manager;
            m_vsq = vsq;
            m_presend = presend;
            m_tracks = new int[] { track };
            m_files = new string[] { file };
            InitializeComponent();
            lblProgress.Text = "1/" + 1;
            progressWhole.Maximum = 1;
            m_partial_mode = true;
            m_start = start;
            m_end = end;
            m_temp_premeasure = temp_premeasure;
            ApplyLanguage();
        }
        
        public FormSynthesize( EditorManager manager, VsqFile vsq, int presend, int[] tracks, string[] files, int end ) {
            m_manager = manager;
            m_vsq = vsq;
            m_presend = presend;
            m_tracks = tracks;
            m_files = files;
            InitializeComponent();
            lblProgress.Text = "1/" + m_tracks.Length;
            progressWhole.Maximum = m_tracks.Length;
            m_partial_mode = false;
            m_end = end;
            ApplyLanguage();
        }

        public void ApplyLanguage() {
            this.Text = _( "Synthesize" );
            lblSynthesizing.Text = _( "now synthesizing..." );
            btnCancel.Text = _( "Cancel" );
        }

        private static string _( string id ) {
            return Messaging.GetMessage( id );
        }

        /// <summary>
        /// レンダリングが完了したトラックのリストを取得します
        /// </summary>
        public int[] Finished {
            get {
                List<int> list = new List<int>();
                for ( int i = 0; i <= m_finished; i++ ) {
                    list.Add( m_tracks[i] );
                }
                return list.ToArray();
            }
        }

        private void FormSynthesize_Load( object sender, EventArgs e ) {
            if ( VSTiProxy.CurrentUser == "" ) {
                VSTiProxy.CurrentUser = m_manager.ID;
                timer.Enabled = true;
                m_rendering_started = true;
                bgWork.RunWorkerAsync();
            } else {
                m_rendering_started = false;
                this.DialogResult = DialogResult.Cancel;
            }
        }

        private void UpdateProgress( int value ) {
            progressWhole.Value = value > progressWhole.Maximum ? progressWhole.Maximum : value;
            lblProgress.Text = value + "/" + m_tracks.Length;
            m_finished = value - 1;
        }

        private void bgWork_DoWork( object sender, DoWorkEventArgs e ) {
            if ( m_partial_mode ) {
                this.Invoke( new BSimpleDelegate<int>( this.UpdateProgress ), new object[] { (object)1 } );
                List<VsqNrpn> nrpn = new List<VsqNrpn>( 
                    m_vsq.GenerateNRPN( 
                        m_vsq.Tracks[m_tracks[0]], 
                        m_presend,
                        m_start,
                        m_end,
                        m_temp_premeasure ) );
                int count = m_vsq.Tracks[m_tracks[0]].Events.Count;
                List<TempoTableEntry> tempo = new List<TempoTableEntry>();
                int start_tempo = m_vsq.GetTempoAt( m_start - m_temp_premeasure );
                tempo.Add( new TempoTableEntry( 0, start_tempo, 0.0 ) );
                for ( int i = 0; i < m_vsq.TempoTable.Count; i++ ) {
                    if ( m_vsq.TempoTable[i].Clock > m_end ) {
                        break;
                    }
                    if ( m_vsq.TempoTable[i].Clock >= m_start - m_temp_premeasure ) {
                        tempo.Add( new TempoTableEntry( m_vsq.TempoTable[i].Clock - m_start + m_temp_premeasure, m_vsq.TempoTable[i].Tempo, 0.0 ) );
                    }
                }
                if ( count > 0 ) {
                    VSTiProxy.RenderToWave( VsqNrpn.Convert( nrpn.ToArray() ), tempo.ToArray(), m_end - m_start + m_temp_premeasure, m_files[0] );
                }
            } else {
                for ( int i = 0; i < m_tracks.Length; i++ ) {
                    this.Invoke( new BSimpleDelegate<int>( this.UpdateProgress ), new object[] { (object)(i + 1) } );
                    List<VsqNrpn> nrpn = new List<VsqNrpn>( m_vsq.GenerateNRPN( m_vsq.Tracks[m_tracks[i]], m_presend ) );
                    int count = m_vsq.Tracks[m_tracks[i]].Events.Count;
                    if ( count > 0 ) {
                        VSTiProxy.RenderToWave( VsqNrpn.Convert( nrpn.ToArray() ), m_vsq.TempoTable.ToArray(), m_end, m_files[i] );
                    }
                }
            }
        }

        private void FormSynthesize_FormClosing( object sender, FormClosingEventArgs e ) {
            timer.Enabled = false;
            if ( m_rendering_started ) {
                VSTiProxy.CurrentUser = "";
            }
            if ( bgWork.IsBusy ) {
                VSTiProxy.AbortRendering();
                bgWork.CancelAsync();
                while ( bgWork.IsBusy ) {
                    Application.DoEvents();
                }
                DialogResult = DialogResult.Cancel;
            } else {
                DialogResult = DialogResult.OK;
            }
        }

        private void bgWork_RunWorkerCompleted( object sender, RunWorkerCompletedEventArgs e ) {
            timer.Enabled = false;
            DialogResult = DialogResult.OK;
        }

        private void timer_Tick( object sender, EventArgs e ) {
            double progress = (int)VSTiProxy.GetProgress();
            progressOne.Value = (int)progress > 100 ? 100 : (int)progress;
        }
    }

}
